import websockets
import json
import regex
import string
import random
import asyncio
from .schemas import *

json_rgx = regex.compile(r'''
(?(DEFINE)
(?P<json>(\s*(?&object)\s*|\s*(?&array)\s*))
(?P<object>(\{\s*((?&pair)(\s*,\s*(?&pair))*)?\s*\}))
(?P<pair>((?&STRING)\s*:\s*(?&value)))
(?P<array>(\[\s*((?&value)(\s*,\s*(?&value))*)?\s*\]))
(?P<value>(true|false|null|(?&STRING)|(?&NUMBER)|(?&object)|(?&array)))
(?P<STRING>("(\\?!(["\\bfnrt]|u[a-fA-F0-9]{4})|[^"\0-\x1F\x7F]+)*"))
(?P<NUMBER>(\-?(0|[1-9][0-9]*)(\.[0-9]+)?([eE][+-]?[0-9]+)?))
)
(?&json)
''', regex.VERBOSE | regex.MULTILINE)

class TradingviewBase:
    websocket_url = 'wss://data.tradingview.com/socket.io/websocket'

class TradingViewWsMessages:
    
    def create_message(self, func_name, param_list):
        return self.prepend_header(self.construct_message(func_name, param_list))
    
    def prepend_header(self, st):
        return "~m~" + str(len(st)) + "~m~" + st

    def construct_message(self, func_name, param_list):
        return json.dumps({
            "m":func_name,
            "p":param_list
            }, separators=(',', ':'))
    
    def set_auth_token(self, token):
        message = self.create_message('set_auth_token',[token])
        return message

    def chart_create_session(self, chart_session_string):
        message = self.create_message('chart_create_session',[chart_session_string, ""])
        return message

    def resolve_symbol(self, chart_session_string, tradingview_symbol):#
        symbol_param  = f'={json.dumps({"symbol":tradingview_symbol, "adjustment":"splits"})}'.replace('"','\"')
        message = self.create_message('resolve_symbol', [chart_session_string, 'sds_sym_1', symbol_param])#TODO maybe you'd make symbol_{index} and use it somehow
        return message
    
    def create_series(self, chart_session_string, tradingview_timeframe, limit=5000):
        message = self.create_message('create_series', [chart_session_string, "sds_1", "s1", "sds_sym_1", tradingview_timeframe, 300, "12M"])#TODO map arguments insetad of hardcode
        return message
    
    def quote_fast_symbols(self, quote_session_string, tradingview_symbol):
        message = self.create_message('quote_fast_symbols', [quote_session_string,tradingview_symbol])
        return message
    
    def set_data_quality(self, quality):
        message = self.create_message('set_data_quality', [quality])
        return message
    
    def quote_create_session(self, quote_session_string):
        message = self.create_message('quote_create_session', [quote_session_string])
        return message
    
    def quote_add_symbols(self, quote_session_string, tradingview_symbol):
        message = self.create_message('quote_add_symbols', [quote_session_string,tradingview_symbol])
        return message
    

class TradingviewWsScraper(TradingviewBase):

    headers =  {
        # 'Connection': 'upgrade',
        # 'Host': 'data.tradingview.com',
        'Origin': 'https://data.tradingview.com'
        # 'Cache-Control': 'no-cache',
        # 'Upgrade': 'websocket',
        # 'Sec-WebSocket-Extensions': 'permessage-deflate; client_max_window_bits',
        # 'Sec-WebSocket-Key': '2C08Ri6FwFQw2p4198F/TA==',
        # 'Sec-WebSocket-Version': '13',
        # 'User-Agent': 'User-Agent: Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.116 Safari/537.36 Edg/83.0.478.56',
        # 'Pragma': 'no-cache',
        # 'Upgrade': 'websocket'
    }


    def __init__(self, token = 'unauthorized_user_token') -> None:
        self.token = token
        self.messages = TradingViewWsMessages()
        self.queue = asyncio.Queue()
    
    def _generate_session(self, prefix = 'qs_'):
        string_length=12
        letters = string.ascii_lowercase
        digits = string.digits
        random_string= ''.join(random.choice(letters+digits) for i in range(string_length))
        return prefix + random_string

    def chart_session(self):
        return self._generate_session(prefix='cs_')
    
    def quote_session(self):
        return self._generate_session(prefix='qs_')
        
    def message_handler(self, message: str):
        return [json.loads(i) for i in json_rgx.findall(message.replace('\\"',''))]
        


    async def get_candles(self, tradingview_symbol:str, tradingview_timeframe):
        chart_session_string = self.chart_session()

        messages = [self.messages.set_auth_token(self.token), 
                    self.messages.set_data_quality('low'),
                    self.messages.chart_create_session(chart_session_string), 
                    self.messages.resolve_symbol(chart_session_string, tradingview_symbol),
                    self.messages.create_series(chart_session_string,
                     tradingview_timeframe)]
        
        async with websockets.connect(self.websocket_url, extra_headers=self.headers) as connection:#TODO make a generic iterator
            for message in messages:#TODO wrap it somehow to a separate method sending and receiving messages
                await connection.send(message)
            while True:
                response = await connection.recv()
                for item in self.message_handler(response):
                    try:
                        timescale_update = factory.load(item, TimescaleUpdateMessage)
                        return timescale_update.symbol_data_series.ohlcv
                    except (TypeError, ValueError):
                        await asyncio.sleep(0)
                        continue#XXX play with exceptions a bit
    
    async def get_symbol(self, tradingview_symbol:str):#TODO add bonds support
        """
        retrieve symbol info (descriptions, currency, etc)
        """
        quote_session_string = self.quote_session()

        messages = [self.messages.set_auth_token(self.token), 
                    self.messages.set_data_quality('low'),
                    self.messages.quote_create_session(quote_session_string),
                    self.messages.quote_add_symbols(quote_session_string, tradingview_symbol),
                    self.messages.quote_fast_symbols(quote_session_string, tradingview_symbol)]
        
        async with websockets.connect(self.websocket_url, extra_headers=self.headers) as connection: #TODO make a generic iterator
            for message in messages:#
                await connection.send(message)
            while True:
                response = await connection.recv()
                for item in self.message_handler(response):
                    try:
                        obj = factory.load(item, QuoteSeriesData)
                        return obj.symbol_info
                    except (TypeError, ValueError):
                        await asyncio.sleep(0)
                        continue#XXX play with exceptions a bit
        


class Intervals:
    interval_1minute = '1'
    interval_5minutes = '5'
    interval_1hour = '60'
    interval_1day = 'D'
    
class Sectors:
    cfd = 'cfd'
    america = 'america'
    iceland = 'iceland'
    japan = 'japan'
    luxembourg = 'luxembourg'
    uk = 'uk'
    russia = 'russia'
    australia = 'australia'
    bahrain = 'bahrain'
    belgium = 'belgium'
    brazil = 'brazil'
    vietnam = 'vietnam'
    germany = 'germany'
    hongkong = 'hongkong'
    denmark = 'denmark'
    egypt = 'egypt'
    israel = 'israel'
    india = 'india'
    indonesia = 'indonesia'
    spain = 'spain'
    italy = 'italy'
    canada = 'canada'
    qatar = 'qatar'
    china = 'china'
    colombia = 'colombia'
    latvia = 'latvia'
    lithuania = 'lithuania'
    malaysia = 'malaysia'
    mexico = 'mexico'
    nigeria = 'nigeria'
    netherlands = 'netherlands'
    newzealand = 'newzealand'
    norway = 'norway'
    uae = 'uae'
    peru = 'peru'
    poland = 'poland'
    portugal = 'portugal'
    ksa = 'ksa'
    serbia = 'serbia'
    singapore = 'singapore'
    thailand = 'thailand'
    taiwan = 'taiwan'
    turkey = 'turkey'
    finland = 'finland'
    france = 'france'
    chile = 'chile'
    switzerland = 'switzerland'
    sweden = 'sweden'
    estonia = 'estonia'
    rsa = 'rsa'
    korea = 'korea'


import uplink

# Constants
SCANNER_API = 'https://scanner.tradingview.com/'
SYMBOL_SEARCH_API = 'https://symbol-search.tradingview.com'
headers = {'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/90.0.4430.72 Safari/537.36'}
@uplink.headers(headers)
class Scanner(uplink.Consumer):
    @uplink.returns.json
    @uplink.get("/{sector}/scan")
    def get_sector_symbols(self,sector):
        """Gets tradingview symbols from specified sector"""
        
@uplink.headers(headers)
class SymbolSearch(uplink.Consumer):
    @uplink.returns.json
    @uplink.params({"hl": True, 'lang': "ru", 'domain':'production'})#static params
    @uplink.get('/symbol_search/')
    def search_symbol(self, symbol: uplink.Query("text")):
        """Search symbol in tradingview api"""

SymbolScanner =  Scanner(SCANNER_API)
SymbolSearcher = SymbolSearch(SYMBOL_SEARCH_API)


from dataclasses import InitVar
from dataclasses import dataclass
from typing import Literal, List, Tuple, Optional
from inflection import underscore

import dataclass_factory
from dataclass_factory import Factory, Schema, NameStyle


default_name_mapping = {"message":'m','timestamp':'t','timestamp_milliseconds':'t_ms'}

@dataclass
class InitialResponse:
    session_id: str
    timestamp: int
    timestamp_ms: int
    release : str
    studies_metadata_hash: str
    protocol: str
    javastudies: str
    auth_scheme_vsn: int
    via: str

initial_response_schema = Schema(
    name_mapping={
        'timestamp_ms':'timestampMs'
    }
)

@dataclass
class SeriesLoadingMessage:
    message: Literal['series_loading']
    chart_session: Optional[str] = None
    timestamp: int = None
    timestamp_milliseconds: int = None

series_loading_schema = Schema(
    name_mapping={
        **default_name_mapping,
        "chart_session": ("p", 0)
    }
)

@dataclass
class SymbolResolvedMessage:
    message: Literal['symbol_resolved']
    chart_session: str = None
    timestamp: int = None
    timestamp_milliseconds: int = None
    symbol_info : dict = None
    
symbol_resolved_schema = Schema(
    name_mapping={
        **default_name_mapping,
        "chart_session": ("p", 0),
        "symbol_info":("p",2)
    }
)

@dataclass 
class SeriesTimeframeMessage:#TODO identify what other fields mean. but now theyre worthless ['cs_mmnwr6m6ed90', 'sds_1', 's1', 0, 365, '12M', False] 
    message: Literal['series_timeframe']
    chart_session: str = None

series_timeframe_schema = series_loading_schema

@dataclass
class Ohlcv:
    timestamp: int
    open: float
    high: float
    low: float
    close: float
    volume: float


@dataclass
class SymbolDataSeries:
    node: str = None
    ohlcv: List[Ohlcv] = None

ohlcv_schema = Schema(
    name_mapping={
        'timestamp':('v',0),
        'open':('v',1),
        'high':('v',2),
        'low':('v',3),
        'close':('v',4),
        'volume':('v',5)
    }
)

@dataclass 
class TimescaleUpdateMessage:
    message: Literal['timescale_update']
    chart_session: str = None
    symbol_data_series: SymbolDataSeries = None

symbol_data_series_schema = Schema(
    name_mapping = {
        'ohlcv':'s'
    }
)

timescale_update_schema = Schema(
    name_mapping={
        **default_name_mapping,
        "chart_session": ("p", 0),
        "symbol_data_series":("p",1,'sds_1')
    }
)

@dataclass
class SeriesCompletedMessage:#['cs_vqn2b9kbmahp', 'sds_1', 'streaming', 's1']#
    message: Literal['series_completed']

series_completed_schema = Schema(
    name_mapping={
        **default_name_mapping
    }
)

####
@dataclass
class SymbolInfo:#TODO more fields
    country: str
    country_fund: str
    exchange_traded: str
    industry: str
    business_description: str
    local_description: str
    short_description: str
    group: str
    mic: str
    sector: str
    web_site_url: str
    ceo: str
    location: str
    dividend_type_recent: str
    fundamental_currency_code: str
    symbol_proname: str
    rt_update_time: str
    exchange_listed: str
    exchange_ticker: str
    exchange_listed_symbol: str
    currency_id: str

symbol_info_schema = Schema(
    name_mapping = {underscore(k):k for k in ['exchange-traded','local-description','symbol-proname','rt-update-time','exchange-ticker','exchange-listed','exchange-listed-symbol','short-description','currency-id']}
)

@dataclass
class QuoteSeriesData:
    message: Literal['qsd']
    quote_session: str
    symbol_name: str
    symbol_info: SymbolInfo

quote_series_data_schema  = Schema(
    name_mapping = {
        'message':'m',
        'quote_session':('p',0),
        'symbol_name':('p',1,'n'),
        'symbol_info':('p',1,'v')
    }
)
    

factory = dataclass_factory.Factory(schemas={
    InitialResponse:initial_response_schema,
    SeriesLoadingMessage: series_loading_schema, 
    SymbolResolvedMessage: symbol_resolved_schema,
    SeriesTimeframeMessage: series_timeframe_schema,
    TimescaleUpdateMessage: timescale_update_schema,
    SeriesCompletedMessage: series_completed_schema,
    SymbolDataSeries: symbol_data_series_schema,
    Ohlcv: ohlcv_schema,
    QuoteSeriesData: quote_series_data_schema,
    SymbolInfo:symbol_info_schema
    })


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tvw_scraper']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.8.1,<4.0.0',
 'dataclass-factory>=2.14,<3.0',
 'inflection>=0.5.1,<0.6.0',
 'regex>=2022.1.18,<2023.0.0',
 'tenacity>=8.0.1,<9.0.0',
 'uplink>=0.9.7,<0.10.0',
 'websocket-client>=1.2.3,<2.0.0',
 'websockets>=10.2,<11.0']

setup_kwargs = {
    'name': 'tvw-scraper',
    'version': '0.0.13',
    'description': 'Scraping tradingview.com',
    'long_description': "# tvw_scraper\nuse asyncio and websockets to retrieve candles and symbol info from tradingview.com\n\nUsage example:\n\nretrieve symbols to use them later in websocket queries\n```python\nfrom tvw_scraper.rest import SymbolScanner\n\nSymbolScanner.get_sector_symbols(Sectors.russia)\n{'totalCount': 937, 'data': [{'s': 'MOEX:AFKS', 'd': []}, {'s': 'MOEX:JNJ-RM', 'd': []} . . .\n```\n\nretrieve candles and some info from tradingview websocket:\n```python\nimport asyncio\nfrom tvw_scraper.scraper import TradingviewWsScraper\nfrom tvw_scraper.models import Intervals, Sectors\n\nasync def main():\n    result = await asyncio.gather(*[\n        scraper.get_candles('NASDAQ:NVDA',Intervals.interval_1day), \n        scraper.get_candles('NASDAQ:NVDA',Intervals.interval_1hour), \n        scraper.get_symbol('NASDAQ:NVDA')\n        ])\n\n\nasyncio.get_event_loop().run_until_complete(main())\n```",
    'author': 'fip',
    'author_email': 'ogremagi9@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ogremagi4/tvw_scraper',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
